@@ -3,11 +3,14 @@ BEGIN {
   $Constant::Export::Lazy::AUTHORITY = 'cpan:AVAR';
 }
 {
-  $Constant::Export::Lazy::VERSION = '0.12';
+  $Constant::Export::Lazy::VERSION = '0.13';
 }
 use 5.006;
 use strict;
 use warnings;
+use warnings FATAL => "recursion";
+
+our $_CALL_SHOULD_ALIAS_FROM_TO  = {};
 
 sub import {
     my ($class, %args) = @_;
@@ -62,7 +65,7 @@ sub import {
         my (undef, @gimme) = @_;
         my $pkg_importer = caller;
 
-        my $ctx = Constant::Export::Lazy::Ctx->new(
+        my $ctx = bless {
             constants    => $constants,
             pkg_importer => $pkg_importer,
 
@@ -102,7 +105,7 @@ sub import {
             # generated by e.g. constant.pm, we know we've made all
             # the constants in the package to our liking.
             wrap_existing_import => $wrap_existing_import,
-        );
+        } => 'Constant::Export::Lazy::Ctx';
 
         # We've been provided with a callback to be used to munge
         # whatever we actually got provided with in @gimme to a list
@@ -121,8 +124,24 @@ sub import {
         my @leftover_gimme;
         for my $gimme (@gimme) {
             if (exists $constants->{$gimme}) {
+                # We only want to alias constants into the importer's
+                # package if the constant is on the import list, not
+                # if it's just needed within some $ctx->call() when
+                # defining another constant.
+                #
+                # To disambiguate these two cases we maintain a
+                # globally dynamically scoped variable with the
+                # constants that have been requested, and we note
+                # who've they've been requested by.
+                local $_CALL_SHOULD_ALIAS_FROM_TO->{$pkg_importer}->{$gimme} = undef;
+
                 $ctx->call($gimme);
             } elsif ($wrap_existing_import) {
+                # We won't even die on $wrap_existing_import if that
+                # importer doesn't know about this $gimme, but
+                # hopefully they're just about to die with an error
+                # similar to ours if they don't know about the
+                # requested constant.
                 push @leftover_gimme => $gimme;
             } else {
                 die "PANIC: We don't have the constant '$gimme' to export to you";
@@ -177,26 +196,9 @@ sub _normalize_arguments {
     return \%args;
 }
 
-package Constant::Export::Lazy::Ctx;
-BEGIN {
-  $Constant::Export::Lazy::Ctx::AUTHORITY = 'cpan:AVAR';
-}
-{
-  $Constant::Export::Lazy::Ctx::VERSION = '0.12';
-}
-use strict;
-use warnings;
+our $_GETTING_VALUE_FOR_OVERRIDE = {};
 
-sub new {
-    my ($class, %args) = @_;
-
-    bless \%args => $class;
-}
-
-our $CALL_SHOULD_NOT_ALIAS;
-our $GETTING_VALUE_FOR_OVERRIDE = {};
-
-sub call {
+sub Constant::Export::Lazy::Ctx::call {
     my ($ctx, $gimme) = @_;
 
     # Unpack our options
@@ -215,13 +217,16 @@ sub call {
     my ($private_name, $glob_name, $alias_as);
     my $make_private_glob_and_alias_name = sub {
         # Checking "exists $constants->{$gimme}" here to avoid
-        # autovivification would be redundant since we'll die before
-        # we get here unless $wrap_existing_import is true above. If
-        # $wrap_existing_import is true we'll have called the import()
-        # we're wrapping if it's a constant we don't know about, or
-        # it's being called from ->call(), in which case
-        # $CALL_SHOULD_NOT_ALIAS is true and we won't be calling this
-        # sub unless $constants->{$gimme} exists. Since this won't
+        # autovivification would be redundant since we won't call this
+        # if $wrap_existing_import is true, otherwise
+        # $constants->{$gimme} is guaranteed to exist. See the
+        # assertion just a few lines above this code.
+        #
+        # If $wrap_existing_import is true and we're handling a
+        # constant we don't know about we'll have called the import()
+        # we're wrapping, or we're being called from ->call(), in
+        # which case we won't be calling this sub unless
+        # $constants->{$gimme} exists.
         $private_name = exists $constants->{$gimme}->{options}->{private_name_munger}
                              ? $constants->{$gimme}->{options}->{private_name_munger}->($gimme)
                              : $gimme;
@@ -251,9 +256,13 @@ sub call {
             $value = $value[0];
         }
     } elsif (do {
+        # Check if this is a constant we've defined already, in which
+        # case we can just return its value.
+        #
         # If we got this far we know we're going to want to call
         # $make_private_glob_and_alias_name->(). It'll also be used by
-        # branches further down.
+        # the "else" branch below if we end up having to define this
+        # constant.
         $make_private_glob_and_alias_name->();
 
         exists $symtab->{$private_name};
@@ -263,12 +272,6 @@ sub call {
         my $symtab_value = $symtab->{$private_name};
         $value = &$symtab_value();
     } else {
-        # We only want to alias constants into the importer's package
-        # if the constant is on the import list, not if it's just
-        # needed within some $ctx->call() when defining another
-        # constant.
-        local $CALL_SHOULD_NOT_ALIAS = 1;
-
         my $override = $constants->{$gimme}->{options}->{override};
         my $stash    = $constants->{$gimme}->{options}->{stash};
 
@@ -281,8 +284,10 @@ sub call {
 
         my @overriden_value;
         my $source;
-        if ($override and not exists $GETTING_VALUE_FOR_OVERRIDE->{$gimme}) {
-            local $GETTING_VALUE_FOR_OVERRIDE->{$gimme} = undef;
+        if ($override and
+            not (exists $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer} and
+                 exists $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer}->{$gimme})) {
+            local $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer}->{$gimme} = undef;
             @overriden_value = $override->($ctx, $gimme);
         }
         if (@overriden_value) {
@@ -299,7 +304,8 @@ sub call {
             $value = $constants->{$gimme}->{call}->($ctx);
         }
 
-        unless (exists $GETTING_VALUE_FOR_OVERRIDE->{$gimme}) {
+        unless (exists $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer} and
+                exists $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer}->{$gimme}) {
             # Instead of doing `sub () { $value }` we could also
             # use the following trick that constant.pm uses if
             # it's true that `$] > 5.009002`:
@@ -375,7 +381,48 @@ sub call {
         }
     }
 
-    unless ($CALL_SHOULD_NOT_ALIAS) {
+    # So? What's this entire evil magic about?
+    #
+    # Early on in the history of this module I decided that everything
+    # that needed to call or define a constant would just go through
+    # $ctx->call($gimme), including things called via the import().
+    #
+    # This makes some parts of this module much simpler, since we
+    # don't have e.g. a $ctx->call_and_intern($gimme) to define
+    # constants for the first time, v.s. a
+    # $ctx->get_interned_value($gimme). We just have one
+    # $ctx->call($gimme) that DWYM. You just request a value, it does
+    # the right thing, and you don't have to worry about it.
+    #
+    # However, we have to worry about the following cases:
+    #
+    # * Someone in "user" imports YourExporter::CONSTANT, we define
+    #   YourExporter::CONSTANT and alias user::CONSTANT to it. Easy,
+    #   this is the common case.
+    #
+    # * Ditto, but YourExporter::CONSTANT needs to get the value of
+    #   YourExporter::CONSTANT_NESTED to define its own value, we want
+    #   to export YourExporter::CONSTANT to user::CONSTANT but *NOT*
+    #   YourExporter::CONSTANT_NESTED. We don't want to leak dependent
+    #   constants like that.
+    #
+    # * The "user" imports YourExporter::CONSTANT, this in turns needs
+    #   to call Some::Module::function() and Some::Module::function()
+    #   needs YourExporter::UNRELATED_CONSTANT
+    #
+    # * When we're in the "override" callback for
+    #   YourExporter::CONSTANT we don't want to intern
+    #   YourExporter::CONSTANT, but if we call some unrelated
+    #   YourExporter::ANOTHER_CONSTANT while in the override we want
+    #   to intern (but not export!) that value.
+    #
+    # So to do all this we're tracking on a per importer/constant pair
+    # basis who requested what during import()-time, and whether we're
+    # currently in the scope of an "override" for a given constant.
+    if (not (exists $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer} and
+             exists $_GETTING_VALUE_FOR_OVERRIDE->{$pkg_importer}->{$gimme}) and
+        exists $_CALL_SHOULD_ALIAS_FROM_TO->{$pkg_importer} and
+        exists $_CALL_SHOULD_ALIAS_FROM_TO->{$pkg_importer}->{$gimme}) {
         no strict 'refs';
         # Alias e.g. user::CONSTANT to YourExporter::CONSTANT (well,
         # actually YourExporter::$private_name)
@@ -385,12 +432,12 @@ sub call {
     return $value;
 }
 
-sub stash {
+sub Constant::Export::Lazy::Ctx::stash {
     my ($ctx) = @_;
 
-    # We used to die here but that makes e.g. having a global "after"
-    # callback tedious. Just return an empty list instead so we can do
-    # things like:
+    # We used to die here when no $ctx->{stash} existed, but that
+    # makes e.g. having a global "after" callback tedious. Just return
+    # undef instead so we can do things like:
     #
     #    if (defined(my $stash = $ctx->stash)) { ... }
     #
@@ -409,9 +456,10 @@ Constant::Export::Lazy - Utility to write lazy exporters of constant subroutines
 
 =head1 SYNOPSIS
 
-This increasingly verbose example of a C<My::Constants> package that
-you can write using C<Constant::Export::Lazy> demonstrates most of our
-major features (from F<t/lib/My/Constants.pm> in the source distro):
+This as an example of a C<My::Constants> package that you can write
+using C<Constant::Export::Lazy> that demonstrates most of its main
+features. This is from the file F<t/lib/My/Constants.pm> in the source
+distro:
 
     package My::Constants;
     use strict;
@@ -662,7 +710,7 @@ And running it gives:
 
 =head1 DESCRIPTION
 
-This is a utility to write lazy exporters of constant
+This is a library to write lazy exporters of constant
 subroutines. It's not meant to be a user-facing constant exporting
 API, it's something you use to write user-facing constant exporting
 APIs.
@@ -672,10 +720,10 @@ or another, why did I need to write this one?
 
 =head2 It's lazy
 
-Our constants fleshened via callbacks that are guaranteed to be called
-only once for the lifetime of the process (not once per importer or
-whatever), and we only call the callbacks lazily if someone actually
-requests that a constant of ours be defined.
+Our constants are fleshened via callbacks that are guaranteed to be
+called only once for the lifetime of the process (not once per
+importer or whatever), and we only call the callbacks lazily if
+someone actually requests that a constant of ours be defined.
 
 This makes it easy to have one constant exporting module that runs in
 different environments, and generates some subset of its constants
@@ -687,8 +735,8 @@ available everywhere, or make certain assumptions about the
 environment they're running under that may not be true across all your
 environments.
 
-By only defining those constants you actually need via callbacks that
-you provide managing all these special-cases becomes a lot easier.
+By only defining those constants you actually need via callbacks
+managing all these special-cases becomes a lot easier.
 
 =head2 It makes it easier to manage creating constants that require other constants
 
@@ -700,17 +748,19 @@ Now say you have several hundred constants like that, managing the
 inter-dependencies and ensuring that they're all defined in the right
 order with dependencies before dependents quickly gets messy.
 
-Constant::Import::Lazy takes away all this complexity. When you define
-a constant you get a callback object that can give you the value of
-other constants.
+All this complexity becomes a non-issue when you use this module. When
+you define a constant you get a callback object that can give you the
+value of other constants.
 
 When you look up another constant we'll either generate it if it
-hasn't been materialized yet, look it up in the symbol table if it
-has.
+hasn't been materialized yet, or look up the materialized value in the
+symbol table if it has.
 
 Thus we end up with a Makefile-like system where you can freely use
-whatever other constants you like when defining your constants, you
-only have to be careful not to introduce circular dependencies.
+whatever other constants you like when defining your constants, and
+we'll lazily define the entire tree of constants on-demand.
+
+You only have to be careful not to introduce circular dependencies.
 
 =head1 API
 
@@ -759,9 +809,9 @@ L<synopsis|/SYNOPSIS>) to strip or map parameters to e.g. implement
 support for C<%EXPORT_TAGS>, or to do any other arbitrary mapping.
 
 This callback will be called with a reference to the parameters passed
-to import, and for convenience with the L<constants|/constants> with
-the hash you provided (e.g. for introspecting the stashes of
-constants, see the L<synopsis|/SYNOPSIS> example.
+to import, and for convenience with the L<constants hash|/constants>
+you provided (e.g. for introspecting the stashes of constants, see the
+L<synopsis|/SYNOPSIS> example.
 
 This is expected to return an array with a list of constants to
 import, or the empty list if we should discard the return value of
@@ -798,7 +848,8 @@ won't have to change any of the old code.
 We'll handle calling subroutines generated with perl's own
 L<constant.pm|constant> (including "list" constants), but we'll die in
 C<call> if we call a foreign subroutine that returns more than one
-value.
+value, i.e. constants defined as C<use constant FOO => (1, 2, 3)>
+instead of C<use constant FOO => [1, 2, 3]>.
 
 =head3 override
 
